@@ -22,16 +22,19 @@
 
 struct StringIO {
     VALUE string;
+    rb_encoding *enc;
     long pos;
     long lineno;
     int flags;
     int count;
 };
 
-static void strio_init(int, VALUE *, struct StringIO *, VALUE);
+static VALUE strio_init(int, VALUE *, struct StringIO *, VALUE);
+static VALUE strio_unget_bytes(struct StringIO *, const char *, long);
 
 #define IS_STRIO(obj) (rb_typeddata_is_kind_of((obj), &strio_data_type))
-#define error_inval(msg) (errno = EINVAL, rb_sys_fail(msg))
+#define error_inval(msg) (rb_syserr_fail(EINVAL, msg))
+#define get_enc(ptr) ((ptr)->enc ? (ptr)->enc : rb_enc_get((ptr)->string))
 
 static struct StringIO *
 strio_alloc(void)
@@ -92,17 +95,25 @@ get_strio(VALUE self)
     return ptr;
 }
 
+static VALUE
+enc_subseq(VALUE str, long pos, long len, rb_encoding *enc)
+{
+    str = rb_str_subseq(str, pos, len);
+    rb_enc_associate(str, enc);
+    return str;
+}
+
 static VALUE
 strio_substr(struct StringIO *ptr, long pos, long len)
 {
     VALUE str = ptr->string;
-    rb_encoding *enc = rb_enc_get(str);
+    rb_encoding *enc = get_enc(ptr);
     long rlen = RSTRING_LEN(str) - pos;
 
     if (len > rlen) len = rlen;
     if (len < 0) len = 0;
     if (len == 0) return rb_str_new(0,0);
-    return rb_enc_str_new(RSTRING_PTR(str)+pos, len, enc);
+    return enc_subseq(str, pos, len, enc);
 }
 
 #define StringIO(obj) get_strio(obj)
@@ -168,11 +179,10 @@ strio_initialize(int argc, VALUE *argv, VALUE self)
 	DATA_PTR(self) = ptr = strio_alloc();
     }
     rb_call_super(0, 0);
-    strio_init(argc, argv, ptr, self);
-    return self;
+    return strio_init(argc, argv, ptr, self);
 }
 
-static void
+static VALUE
 strio_init(int argc, VALUE *argv, struct StringIO *ptr, VALUE self)
 {
     VALUE string, mode;
@@ -192,8 +202,7 @@ strio_init(int argc, VALUE *argv, struct StringIO *ptr, VALUE self)
 	}
 	StringValue(string);
 	if ((ptr->flags & FMODE_WRITABLE) && OBJ_FROZEN(string)) {
-	    errno = EACCES;
-	    rb_sys_fail(0);
+	    rb_syserr_fail(EACCES, 0);
 	}
 	if (trunc) {
 	    rb_str_resize(string, 0);
@@ -209,9 +218,11 @@ strio_init(int argc, VALUE *argv, struct StringIO *ptr, VALUE self)
 	break;
     }
     ptr->string = string;
+    ptr->enc = 0;
     ptr->pos = 0;
     ptr->lineno = 0;
     RBASIC(self)->flags |= (ptr->flags & FMODE_READWRITE) * (STRIO_READABLE / FMODE_READABLE);
+    return self;
 }
 
 static VALUE
@@ -238,6 +249,19 @@ strio_s_open(int argc, VALUE *argv, VALUE klass)
     return rb_ensure(rb_yield, obj, strio_finalize, obj);
 }
 
+/* :nodoc: */
+static VALUE
+strio_s_new(int argc, VALUE *argv, VALUE klass)
+{
+    if (rb_block_given_p()) {
+	VALUE cname = rb_obj_as_string(klass);
+
+	rb_warn("%"PRIsVALUE"::new() does not take block; use %"PRIsVALUE"::open() instead",
+		cname, cname);
+    }
+    return rb_class_new_instance(argc, argv, klass);
+}
+
 /*
  * Returns +false+.  Just for compatibility to IO.
  */
@@ -490,7 +514,18 @@ strio_set_lineno(VALUE self, VALUE lineno)
     return lineno;
 }
 
-#define strio_binmode strio_self
+static VALUE
+strio_binmode(VALUE self)
+{
+    struct StringIO *ptr = StringIO(self);
+    rb_encoding *enc = rb_ascii8bit_encoding();
+
+    ptr->enc = enc;
+    if (WRITABLE(self)) {
+	rb_enc_associate(ptr->string, enc);
+    }
+    return self;
+}
 
 #define strio_fcntl strio_unimpl
 
@@ -513,8 +548,7 @@ strio_reopen(int argc, VALUE *argv, VALUE self)
     if (argc == 1 && !RB_TYPE_P(*argv, T_STRING)) {
 	return strio_copy(self, *argv);
     }
-    strio_init(argc, argv, StringIO(self), self);
-    return self;
+    return strio_init(argc, argv, StringIO(self), self);
 }
 
 /*
@@ -662,17 +696,19 @@ static VALUE
 strio_getc(VALUE self)
 {
     struct StringIO *ptr = readable(self);
-    rb_encoding *enc = rb_enc_get(ptr->string);
+    rb_encoding *enc = get_enc(ptr);
+    VALUE str = ptr->string;
+    long pos = ptr->pos;
     int len;
     char *p;
 
-    if (ptr->pos >= RSTRING_LEN(ptr->string)) {
+    if (pos >= RSTRING_LEN(str)) {
 	return Qnil;
     }
-    p = RSTRING_PTR(ptr->string)+ptr->pos;
-    len = rb_enc_mbclen(p, RSTRING_END(ptr->string), enc);
+    p = RSTRING_PTR(str)+pos;
+    len = rb_enc_mbclen(p, RSTRING_END(str), enc);
     ptr->pos += len;
-    return rb_enc_str_new(p, len, rb_enc_get(ptr->string));
+    return enc_subseq(str, pos, len, enc);
 }
 
 /*
@@ -723,19 +759,17 @@ static VALUE
 strio_ungetc(VALUE self, VALUE c)
 {
     struct StringIO *ptr = readable(self);
-    long lpos, clen;
-    char *p, *pend;
     rb_encoding *enc, *enc2;
 
-    if (NIL_P(c)) return Qnil;
     check_modifiable(ptr);
+    if (NIL_P(c)) return Qnil;
     if (FIXNUM_P(c)) {
 	int cc = FIX2INT(c);
 	char buf[16];
 
 	enc = rb_enc_get(ptr->string);
 	rb_enc_mbcput(cc, buf, enc);
-	c = rb_enc_str_new(buf, rb_enc_codelen(cc, enc), enc);
+	return strio_unget_bytes(ptr, buf, rb_enc_codelen(cc, enc));
     }
     else {
 	SafeStringValue(c);
@@ -744,29 +778,10 @@ strio_ungetc(VALUE self, VALUE c)
 	if (enc != enc2 && enc != rb_ascii8bit_encoding()) {
 	    c = rb_str_conv_enc(c, enc2, enc);
 	}
+	strio_unget_bytes(ptr, RSTRING_PTR(c), RSTRING_LEN(c));
+	RB_GC_GUARD(c);
+	return Qnil;
     }
-    if (RSTRING_LEN(ptr->string) < ptr->pos) {
-	long len = RSTRING_LEN(ptr->string);
-	rb_str_resize(ptr->string, ptr->pos - 1);
-	memset(RSTRING_PTR(ptr->string) + len, 0, ptr->pos - len - 1);
-	rb_str_concat(ptr->string, c);
-	ptr->pos--;
-    }
-    else {
-	/* get logical position */
-	lpos = 0; p = RSTRING_PTR(ptr->string); pend = p + ptr->pos;
-	for (;;) {
-	    clen = rb_enc_mbclen(p, pend, enc);
-	    if (p+clen >= pend) break;
-	    p += clen;
-	    lpos++;
-	}
-	clen = p - RSTRING_PTR(ptr->string);
-	rb_str_update(ptr->string, lpos, ptr->pos ? 1 : 0, c);
-	ptr->pos = clen;
-    }
-
-    return Qnil;
 }
 
 /*
@@ -780,35 +795,53 @@ strio_ungetbyte(VALUE self, VALUE c)
 {
     struct StringIO *ptr = readable(self);
     char buf[1], *cp = buf;
-    long pos = ptr->pos, cl = 1;
-    VALUE str = ptr->string;
+    long cl = 1;
 
+    check_modifiable(ptr);
     if (NIL_P(c)) return Qnil;
     if (FIXNUM_P(c)) {
 	buf[0] = (char)FIX2INT(c);
+	return strio_unget_bytes(ptr, buf, 1);
     }
     else {
 	SafeStringValue(c);
 	cp = RSTRING_PTR(c);
 	cl = RSTRING_LEN(c);
 	if (cl == 0) return Qnil;
+	strio_unget_bytes(ptr, cp, cl);
+	RB_GC_GUARD(c);
+	return Qnil;
     }
-    check_modifiable(ptr);
-    rb_str_modify(str);
+}
+
+static VALUE
+strio_unget_bytes(struct StringIO *ptr, const char *cp, long cl)
+{
+    long pos = ptr->pos, len, rest;
+    VALUE str = ptr->string;
+    char *s;
+
+    len = RSTRING_LEN(str);
+    rest = pos - len;
     if (cl > pos) {
-	char *s;
-	long rest = RSTRING_LEN(str) - pos;
-	rb_str_resize(str, rest + cl);
+	long ex = (rest < 0 ? cl-pos : cl+rest);
+	rb_str_modify_expand(str, ex);
+	rb_str_set_len(str, len + ex);
 	s = RSTRING_PTR(str);
-	memmove(s + cl, s + pos, rest);
+	if (rest < 0) memmove(s + cl, s + pos, -rest);
 	pos = 0;
     }
     else {
+	if (rest > 0) {
+	    rb_str_modify_expand(str, rest);
+	    rb_str_set_len(str, len + rest);
+	}
+	s = RSTRING_PTR(str);
+	if (rest > cl) memset(s + len, 0, rest - cl);
 	pos -= cl;
     }
-    memcpy(RSTRING_PTR(str) + pos, cp, cl);
+    memcpy(s + pos, cp, cl);
     ptr->pos = pos;
-    RB_GC_GUARD(c);
     return Qnil;
 }
 
@@ -890,7 +923,7 @@ strio_each_codepoint(VALUE self)
     RETURN_ENUMERATOR(self, 0, 0);
 
     ptr = readable(self);
-    enc = rb_enc_get(ptr->string);
+    enc = get_enc(ptr);
     for (;;) {
 	if (ptr->pos >= RSTRING_LEN(ptr->string)) {
 	    return self;
@@ -949,12 +982,16 @@ bm_search(const char *little, long llen, const char *big, long blen, const long
     return -1;
 }
 
-static VALUE
-strio_getline(int argc, VALUE *argv, struct StringIO *ptr)
+struct getline_arg {
+    VALUE rs;
+    long limit;
+};
+
+static struct getline_arg *
+prepare_getline_args(struct getline_arg *arg, int argc, VALUE *argv)
 {
-    const char *s, *e, *p;
-    long n, limit = 0;
     VALUE str, lim;
+    long limit = -1;
 
     rb_scan_args(argc, argv, "02", &str, &lim);
     switch (argc) {
@@ -967,7 +1004,6 @@ strio_getline(int argc, VALUE *argv, struct StringIO *ptr)
 	    VALUE tmp = rb_check_string_type(str);
 	    if (NIL_P(tmp)) {
 		limit = NUM2LONG(str);
-		if (limit == 0) return rb_str_new(0,0);
 		str = rb_rs;
 	    }
 	    else {
@@ -981,6 +1017,17 @@ strio_getline(int argc, VALUE *argv, struct StringIO *ptr)
 	if (!NIL_P(lim)) limit = NUM2LONG(lim);
 	break;
     }
+    arg->rs = str;
+    arg->limit = limit;
+    return arg;
+}
+
+static VALUE
+strio_getline(struct getline_arg *arg, struct StringIO *ptr)
+{
+    const char *s, *e, *p;
+    long n, limit = arg->limit;
+    VALUE str = arg->rs;
 
     if (ptr->pos >= (n = RSTRING_LEN(ptr->string))) {
 	return Qnil;
@@ -988,8 +1035,8 @@ strio_getline(int argc, VALUE *argv, struct StringIO *ptr)
     s = RSTRING_PTR(ptr->string);
     e = s + RSTRING_LEN(ptr->string);
     s += ptr->pos;
-    if (limit > 0 && s + limit < e) {
-	e = rb_enc_right_char_head(s, s + limit, e, rb_enc_get(ptr->string));
+    if (limit > 0 && (size_t)limit < (size_t)(e - s)) {
+	e = rb_enc_right_char_head(s, s + limit, e, get_enc(ptr));
     }
     if (NIL_P(str)) {
 	str = strio_substr(ptr, ptr->pos, e - s);
@@ -1053,8 +1100,14 @@ strio_getline(int argc, VALUE *argv, struct StringIO *ptr)
 static VALUE
 strio_gets(int argc, VALUE *argv, VALUE self)
 {
-    VALUE str = strio_getline(argc, argv, readable(self));
+    struct getline_arg arg;
+    VALUE str;
 
+    if (prepare_getline_args(&arg, argc, argv)->limit == 0) {
+	return rb_str_new(0, 0);
+    }
+
+    str = strio_getline(&arg, readable(self));
     rb_lastline_set(str);
     return str;
 }
@@ -1093,16 +1146,16 @@ static VALUE
 strio_each(int argc, VALUE *argv, VALUE self)
 {
     VALUE line;
+    struct getline_arg arg;
 
     StringIO(self);
     RETURN_ENUMERATOR(self, argc, argv);
 
-    if (argc > 0 && !NIL_P(argv[argc-1]) && NIL_P(rb_check_string_type(argv[argc-1])) &&
-	NUM2LONG(argv[argc-1]) == 0) {
+    if (prepare_getline_args(&arg, argc, argv)->limit == 0) {
 	rb_raise(rb_eArgError, "invalid limit: 0 for each_line");
     }
 
-    while (!NIL_P(line = strio_getline(argc, argv, readable(self)))) {
+    while (!NIL_P(line = strio_getline(&arg, readable(self)))) {
 	rb_yield(line);
     }
     return self;
@@ -1132,15 +1185,15 @@ static VALUE
 strio_readlines(int argc, VALUE *argv, VALUE self)
 {
     VALUE ary, line;
+    struct getline_arg arg;
 
     StringIO(self);
     ary = rb_ary_new();
-    if (argc > 0 && !NIL_P(argv[argc-1]) && NIL_P(rb_check_string_type(argv[argc-1])) &&
-	NUM2LONG(argv[argc-1]) == 0) {
+    if (prepare_getline_args(&arg, argc, argv)->limit == 0) {
 	rb_raise(rb_eArgError, "invalid limit: 0 for readlines");
     }
 
-    while (!NIL_P(line = strio_getline(argc, argv, readable(self)))) {
+    while (!NIL_P(line = strio_getline(&arg, readable(self)))) {
 	rb_ary_push(ary, line);
     }
     return ary;
@@ -1166,7 +1219,7 @@ strio_write(VALUE self, VALUE str)
 
     if (!RB_TYPE_P(str, T_STRING))
 	str = rb_obj_as_string(str);
-    enc = rb_enc_get(ptr->string);
+    enc = get_enc(ptr);
     enc2 = rb_enc_get(str);
     if (enc != enc2 && enc != ascii8bit) {
 	str = rb_str_conv_enc(str, enc2, enc);
@@ -1441,7 +1494,8 @@ strio_truncate(VALUE self, VALUE len)
 static VALUE
 strio_external_encoding(VALUE self)
 {
-    return rb_enc_from_encoding(rb_enc_get(StringIO(self)->string));
+    struct StringIO *ptr = StringIO(self);
+    return rb_enc_from_encoding(get_enc(ptr));
 }
 
 /*
@@ -1472,7 +1526,7 @@ static VALUE
 strio_set_encoding(int argc, VALUE *argv, VALUE self)
 {
     rb_encoding* enc;
-    VALUE str = StringIO(self)->string;
+    struct StringIO *ptr = StringIO(self);
     VALUE ext_enc, int_enc, opt;
 
     argc = rb_scan_args(argc, argv, "11:", &ext_enc, &int_enc, &opt);
@@ -1483,7 +1537,11 @@ strio_set_encoding(int argc, VALUE *argv, VALUE self)
     else {
 	enc = rb_to_encoding(ext_enc);
     }
-    rb_enc_associate(str, enc);
+    ptr->enc = enc;
+    if (WRITABLE(self)) {
+	rb_enc_associate(ptr->string, enc);
+    }
+
     return self;
 }
 
@@ -1507,6 +1565,7 @@ Init_stringio(void)
 
     rb_include_module(StringIO, rb_mEnumerable);
     rb_define_alloc_func(StringIO, strio_s_allocate);
+    rb_define_singleton_method(StringIO, "new", strio_s_new, -1);
     rb_define_singleton_method(StringIO, "open", strio_s_open, -1);
     rb_define_method(StringIO, "initialize", strio_initialize, -1);
     rb_define_method(StringIO, "initialize_copy", strio_copy, 1);
