@@ -2,7 +2,7 @@
 
   object.c -
 
-  $Author: naruse $
+  $Author: nobu $
   created at: Thu Jul 15 12:01:24 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -145,7 +145,7 @@ rb_obj_equal(VALUE obj1, VALUE obj2)
 #if 0
 /*
  * call-seq:
- *    obj.hash    -> fixnum
+ *    obj.hash    -> integer
  *
  * Generates a Fixnum hash value for this object.  This function must have the
  * property that <code>a.eql?(b)</code> implies <code>a.hash == b.hash</code>.
@@ -268,7 +268,7 @@ rb_obj_copy_ivar(VALUE dest, VALUE obj)
 	RBASIC(dest)->flags |= ROBJECT_EMBED;
     }
     else {
-	long len = ROBJECT(obj)->as.heap.numiv;
+	uint32_t len = ROBJECT(obj)->as.heap.numiv;
 	VALUE *ptr = 0;
 	if (len > 0) {
 	    ptr = ALLOC_N(VALUE, len);
@@ -299,11 +299,12 @@ init_copy(VALUE dest, VALUE obj)
 
 /*
  *  call-seq:
- *     obj.clone -> an_object
+ *     obj.clone(freeze: true) -> an_object
  *
  *  Produces a shallow copy of <i>obj</i>---the instance variables of
  *  <i>obj</i> are copied, but not the objects they reference.
- *  <code>clone</code> copies the frozen and tainted state of <i>obj</i>.
+ *  <code>clone</code> copies the frozen (unless :freeze keyword argument
+ *  is given with a false value) and tainted state of <i>obj</i>.
  *  See also the discussion under <code>Object#dup</code>.
  *
  *     class Klass
@@ -321,11 +322,28 @@ init_copy(VALUE dest, VALUE obj)
  *  the class.
  */
 
-VALUE
-rb_obj_clone(VALUE obj)
+static VALUE
+rb_obj_clone2(int argc, VALUE *argv, VALUE obj)
 {
+    static ID keyword_ids[1];
+    VALUE opt;
+    VALUE kwargs[1];
     VALUE clone;
     VALUE singleton;
+    VALUE kwfreeze = Qtrue;
+
+    if (!keyword_ids[0]) {
+	CONST_ID(keyword_ids[0], "freeze");
+    }
+    rb_scan_args(argc, argv, "0:", &opt);
+    if (!NIL_P(opt)) {
+	rb_get_kwargs(opt, keyword_ids, 0, 1, kwargs);
+	kwfreeze = kwargs[0];
+	if (kwfreeze != Qundef && kwfreeze != Qtrue && kwfreeze != Qfalse) {
+	    rb_raise(rb_eArgError, "unexpected value for freeze: %s",
+		     rb_builtin_class_name(kwfreeze));
+	}
+    }
 
     if (rb_special_const_p(obj)) {
         rb_raise(rb_eTypeError, "can't clone %s", rb_obj_classname(obj));
@@ -342,11 +360,20 @@ rb_obj_clone(VALUE obj)
 
     init_copy(clone, obj);
     rb_funcall(clone, id_init_clone, 1, obj);
-    RBASIC(clone)->flags |= RBASIC(obj)->flags & FL_FREEZE;
+
+    if (Qfalse != kwfreeze) {
+	RBASIC(clone)->flags |= RBASIC(obj)->flags & FL_FREEZE;
+    }
 
     return clone;
 }
 
+VALUE
+rb_obj_clone(VALUE obj)
+{
+    return rb_obj_clone2(0, NULL, obj);
+}
+
 /*
  *  call-seq:
  *     obj.dup -> an_object
@@ -467,22 +494,23 @@ rb_any_to_s(VALUE obj)
 
 VALUE rb_str_escape(VALUE str);
 /*
- * If the default external encoding is ASCII compatible, the encoding of
- * the inspected result must be compatible with it.
- * If the default external encoding is ASCII incompatible,
+ * If the default internal or external encoding is ASCII compatible,
+ * the encoding of the inspected result must be compatible with it.
+ * If the default internal or external encoding is ASCII incompatible,
  * the result must be ASCII only.
  */
 VALUE
 rb_inspect(VALUE obj)
 {
     VALUE str = rb_obj_as_string(rb_funcallv(obj, id_inspect, 0, 0));
-    rb_encoding *ext = rb_default_external_encoding();
-    if (!rb_enc_asciicompat(ext)) {
+    rb_encoding *enc = rb_default_internal_encoding();
+    if (enc == NULL) enc = rb_default_external_encoding();
+    if (!rb_enc_asciicompat(enc)) {
 	if (!rb_enc_str_asciionly_p(str))
 	    return rb_str_escape(str);
 	return str;
     }
-    if (rb_enc_get(str) != ext && !rb_enc_str_asciionly_p(str))
+    if (rb_enc_get(str) != enc && !rb_enc_str_asciionly_p(str))
 	return rb_str_escape(str);
     return str;
 }
@@ -1522,7 +1550,7 @@ rb_mod_freeze(VALUE mod)
  *     mod === obj    -> true or false
  *
  *  Case Equality---Returns <code>true</code> if <i>obj</i> is an
- *  instance of <i>mod</i> or and instance of one of <i>mod</i>'s descendants.
+ *  instance of <i>mod</i> or an instance of one of <i>mod</i>'s descendants.
  *  Of limited use for modules, but can be used in <code>case</code> statements
  *  to classify objects by class.
  */
@@ -1548,18 +1576,15 @@ rb_mod_eqq(VALUE mod, VALUE arg)
 VALUE
 rb_class_inherited_p(VALUE mod, VALUE arg)
 {
-    VALUE start = mod;
-
     if (mod == arg) return Qtrue;
     if (!CLASS_OR_MODULE_P(arg) && !RB_TYPE_P(arg, T_ICLASS)) {
 	rb_raise(rb_eTypeError, "compared with non class/module");
     }
-    arg = RCLASS_ORIGIN(arg);
-    if (class_search_ancestor(mod, arg)) {
+    if (class_search_ancestor(mod, RCLASS_ORIGIN(arg))) {
 	return Qtrue;
     }
     /* not mod < arg; check if mod > arg */
-    if (class_search_ancestor(arg, start)) {
+    if (class_search_ancestor(arg, mod)) {
 	return Qfalse;
     }
     return Qnil;
@@ -1676,7 +1701,7 @@ rb_class_s_alloc(VALUE klass)
  *
  *  Creates a new anonymous module. If a block is given, it is passed
  *  the module object, and the block is evaluated in the context of this
- *  module using <code>module_eval</code>.
+ *  module like <code>module_eval</code>.
  *
  *     fred = Module.new do
  *       def meth1
@@ -1725,7 +1750,7 @@ rb_mod_initialize_clone(VALUE clone, VALUE orig)
  *  class a name by assigning the class object to a constant.
  *
  *  If a block is given, it is passed the class object, and the block
- *  is evaluated in the context of this class using
+ *  is evaluated in the context of this class like
  *  <code>class_eval</code>.
  *
  *     fred = Class.new do
@@ -2141,7 +2166,7 @@ rb_mod_const_get(int argc, VALUE *argv, VALUE mod)
 		continue;
 	    }
 	    else {
-		rb_name_err_raise("uninitialized constant %2$s::%1$s", mod, part);
+		rb_mod_const_missing(mod, part);
 	    }
 	}
 	if (!rb_is_const_id(id)) {
@@ -2179,6 +2204,7 @@ rb_mod_const_set(VALUE mod, VALUE name, VALUE value)
     ID id = id_for_setter(mod, name, const, wrong_constant_name);
     if (!id) id = rb_intern_str(name);
     rb_const_set(mod, id, value);
+
     return value;
 }
 
@@ -2998,7 +3024,7 @@ rb_num_to_dbl(VALUE val)
 {
     if (SPECIAL_CONST_P(val)) {
 	if (FIXNUM_P(val)) {
-	    if (basic_to_f_p(rb_cFixnum))
+	    if (basic_to_f_p(rb_cInteger))
 		return fix2dbl_without_to_f(val);
 	}
 	else if (FLONUM_P(val)) {
@@ -3013,7 +3039,7 @@ rb_num_to_dbl(VALUE val)
 	  case T_FLOAT:
 	    return rb_float_noflonum_value(val);
 	  case T_BIGNUM:
-	    if (basic_to_f_p(rb_cBignum))
+	    if (basic_to_f_p(rb_cInteger))
 		return big2dbl_without_to_f(val);
 	    break;
 	  case T_RATIONAL:
@@ -3169,12 +3195,22 @@ dig_basic_p(VALUE obj, struct dig_method *cache)
     return cache->basic;
 }
 
+static void
+no_dig_method(int found, VALUE recv, ID mid, int argc, const VALUE *argv, VALUE data)
+{
+    if (!found) {
+	rb_raise(rb_eTypeError, "%"PRIsVALUE" does not have #dig method",
+		 CLASS_OF(data));
+    }
+}
+
 VALUE
 rb_obj_dig(int argc, VALUE *argv, VALUE obj, VALUE notfound)
 {
     struct dig_method hash = {Qnil}, ary = {Qnil}, strt = {Qnil};
 
     for (; argc > 0; ++argv, --argc) {
+	if (NIL_P(obj)) return notfound;
 	if (!SPECIAL_CONST_P(obj)) {
 	    switch (BUILTIN_TYPE(obj)) {
 	      case T_HASH:
@@ -3197,7 +3233,8 @@ rb_obj_dig(int argc, VALUE *argv, VALUE obj, VALUE notfound)
 		break;
 	    }
 	}
-	return rb_check_funcall_default(obj, id_dig, argc, argv, notfound);
+	return rb_check_funcall_with_hook(obj, id_dig, argc, argv,
+					  no_dig_method, obj);
     }
     return obj;
 }
@@ -3414,7 +3451,7 @@ InitVM_Object(void)
 
     rb_define_method(rb_mKernel, "class", rb_obj_class, 0);
     rb_define_method(rb_mKernel, "singleton_class", rb_obj_singleton_class, 0);
-    rb_define_method(rb_mKernel, "clone", rb_obj_clone, 0);
+    rb_define_method(rb_mKernel, "clone", rb_obj_clone2, -1);
     rb_define_method(rb_mKernel, "dup", rb_obj_dup, 0);
     rb_define_method(rb_mKernel, "itself", rb_obj_itself, 0);
     rb_define_method(rb_mKernel, "initialize_copy", rb_obj_init_copy, 1);
@@ -3475,9 +3512,10 @@ InitVM_Object(void)
     rb_undef_alloc_func(rb_cNilClass);
     rb_undef_method(CLASS_OF(rb_cNilClass), "new");
     /*
-     * An alias of +nil+
+     * An obsolete alias of +nil+
      */
     rb_define_global_const("NIL", Qnil);
+    rb_deprecate_constant(rb_cObject, "NIL");
 
     rb_define_method(rb_cModule, "freeze", rb_mod_freeze, 0);
     rb_define_method(rb_cModule, "===", rb_mod_eqq, 1);
@@ -3544,7 +3582,7 @@ InitVM_Object(void)
      * Document-class: Data
      *
      * This is a recommended base class for C extensions using Data_Make_Struct
-     * or Data_Wrap_Struct, see README.EXT for details.
+     * or Data_Wrap_Struct, see doc/extension.rdoc for details.
      */
     rb_cData = rb_define_class("Data", rb_cObject);
     rb_undef_alloc_func(rb_cData);
@@ -3559,9 +3597,10 @@ InitVM_Object(void)
     rb_undef_alloc_func(rb_cTrueClass);
     rb_undef_method(CLASS_OF(rb_cTrueClass), "new");
     /*
-     * An alias of +true+
+     * An obsolete alias of +true+
      */
     rb_define_global_const("TRUE", Qtrue);
+    rb_deprecate_constant(rb_cObject, "TRUE");
 
     rb_cFalseClass = rb_define_class("FalseClass", rb_cObject);
     rb_define_method(rb_cFalseClass, "to_s", false_to_s, 0);
@@ -3573,9 +3612,10 @@ InitVM_Object(void)
     rb_undef_alloc_func(rb_cFalseClass);
     rb_undef_method(CLASS_OF(rb_cFalseClass), "new");
     /*
-     * An alias of +false+
+     * An obsolete alias of +false+
      */
     rb_define_global_const("FALSE", Qfalse);
+    rb_deprecate_constant(rb_cObject, "FALSE");
 }
 
 void
