@@ -1,4 +1,5 @@
 # -*- coding: us-ascii -*-
+# frozen_string_literal: false
 require 'test/unit'
 require 'thread'
 
@@ -95,7 +96,7 @@ class TestThread < Test::Unit::TestCase
   end
 
   def test_mutex_synchronize
-    m = Mutex.new
+    m = Thread::Mutex.new
     r = 0
     num_threads = 10
     loop=100
@@ -119,7 +120,7 @@ class TestThread < Test::Unit::TestCase
 
   def test_mutex_synchronize_yields_no_block_params
     bug8097 = '[ruby-core:53424] [Bug #8097]'
-    assert_empty(Mutex.new.synchronize {|*params| break params}, bug8097)
+    assert_empty(Thread::Mutex.new.synchronize {|*params| break params}, bug8097)
   end
 
   def test_local_barrier
@@ -343,6 +344,62 @@ class TestThread < Test::Unit::TestCase
     INPUT
   end
 
+  def test_report_on_exception
+    assert_separately([], <<~"end;") #do
+      q1 = Thread::Queue.new
+      q2 = Thread::Queue.new
+
+      assert_equal(false, Thread.report_on_exception,
+                   "global flags is false by default")
+      assert_equal(false, Thread.current.report_on_exception)
+
+      Thread.current.report_on_exception = true
+      assert_equal(false,
+                   Thread.start {Thread.current.report_on_exception}.value,
+                  "should not inherit from the parent thread")
+
+      assert_warn("", "exception should be ignored silently") {
+        th = Thread.start {
+          q1.push(Thread.current.report_on_exception)
+          raise "report 1"
+        }
+        assert_equal(false, q1.pop)
+        Thread.pass while th.alive?
+      }
+
+      assert_warn(/report 2/, "exception should be reported") {
+        th = Thread.start {
+          q1.push(Thread.current.report_on_exception = true)
+          raise "report 2"
+        }
+        assert_equal(true, q1.pop)
+        Thread.pass while th.alive?
+      }
+
+      assert_equal(false, Thread.report_on_exception)
+      assert_warn("", "the global flag should not affect already started threads") {
+        th = Thread.start {
+          q2.pop
+          q1.push(Thread.current.report_on_exception)
+          raise "report 3"
+        }
+        q2.push(Thread.report_on_exception = true)
+        assert_equal(false, q1.pop)
+        Thread.pass while th.alive?
+      }
+
+      assert_equal(true, Thread.report_on_exception)
+      assert_warn(/report 4/, "should defaults to the global flag at the start") {
+        th = Thread.start {
+          q1.push(Thread.current.report_on_exception)
+          raise "report 4"
+        }
+        assert_equal(true, q1.pop)
+        Thread.pass while th.alive?
+      }
+    end;
+  end
+
   def test_status_and_stop_p
     a = ::Thread.new { raise("die now") }
     b = Thread.new { Thread.stop }
@@ -456,7 +513,7 @@ class TestThread < Test::Unit::TestCase
   end
 
   def test_mutex_deadlock
-    m = Mutex.new
+    m = Thread::Mutex.new
     m.synchronize do
       assert_raise(ThreadError) do
         m.synchronize do
@@ -467,7 +524,7 @@ class TestThread < Test::Unit::TestCase
   end
 
   def test_mutex_interrupt
-    m = Mutex.new
+    m = Thread::Mutex.new
     m.lock
     t = Thread.new do
       m.lock
@@ -479,7 +536,7 @@ class TestThread < Test::Unit::TestCase
   end
 
   def test_mutex_illegal_unlock
-    m = Mutex.new
+    m = Thread::Mutex.new
     m.lock
     assert_raise(ThreadError) do
       Thread.new do
@@ -489,8 +546,8 @@ class TestThread < Test::Unit::TestCase
   end
 
   def test_mutex_fifo_like_lock
-    m1 = Mutex.new
-    m2 = Mutex.new
+    m1 = Thread::Mutex.new
+    m2 = Thread::Mutex.new
     m1.lock
     m2.lock
     m1.unlock
@@ -498,7 +555,7 @@ class TestThread < Test::Unit::TestCase
     assert_equal(false, m1.locked?)
     assert_equal(false, m2.locked?)
 
-    m3 = Mutex.new
+    m3 = Thread::Mutex.new
     m1.lock
     m2.lock
     m3.lock
@@ -511,7 +568,7 @@ class TestThread < Test::Unit::TestCase
   end
 
   def test_mutex_trylock
-    m = Mutex.new
+    m = Thread::Mutex.new
     assert_equal(true, m.try_lock)
     assert_equal(false, m.try_lock, '[ruby-core:20943]')
 
@@ -696,7 +753,7 @@ class TestThread < Test::Unit::TestCase
   end
 
   def test_handle_interrupted?
-    q = Queue.new
+    q = Thread::Queue.new
     Thread.handle_interrupt(RuntimeError => :never){
       done = false
       th = Thread.new{
@@ -727,7 +784,7 @@ class TestThread < Test::Unit::TestCase
   end
 
   def test_thread_timer_and_ensure
-    assert_normal_exit(<<_eom, 'r36492', timeout: 3)
+    assert_normal_exit(<<_eom, 'r36492', timeout: 10)
     flag = false
     t = Thread.new do
       begin
@@ -747,9 +804,24 @@ _eom
   end
 
   def test_uninitialized
-    c = Class.new(Thread)
-    c.class_eval { def initialize; end }
+    c = Class.new(Thread) {def initialize; end}
     assert_raise(ThreadError) { c.new.start }
+
+    bug11959 = '[ruby-core:72732] [Bug #11959]'
+
+    c = Class.new(Thread) {def initialize; exit; end}
+    assert_raise(ThreadError, bug11959) { c.new }
+
+    c = Class.new(Thread) {def initialize; raise; end}
+    assert_raise(ThreadError, bug11959) { c.new }
+
+    c = Class.new(Thread) {
+      def initialize
+        pending = pending_interrupt?
+        super {pending}
+      end
+    }
+    assert_equal(false, c.new.value, bug11959)
   end
 
   def test_backtrace
@@ -830,7 +902,7 @@ _eom
   def test_main_thread_status_at_exit
     assert_in_out_err([], <<-'INPUT', ["false false aborting"], [])
 require 'thread'
-q = Queue.new
+q = Thread::Queue.new
 Thread.new(Thread.current) {|mth|
   begin
     q.push nil
@@ -897,7 +969,7 @@ q.pop
   end
 
   def test_mutex_owned
-    mutex = Mutex.new
+    mutex = Thread::Mutex.new
 
     assert_equal(mutex.owned?, false)
     mutex.synchronize {
@@ -909,7 +981,7 @@ q.pop
 
   def test_mutex_owned2
     begin
-      mutex = Mutex.new
+      mutex = Thread::Mutex.new
       th = Thread.new {
         # lock forever
         mutex.lock
@@ -926,7 +998,7 @@ q.pop
 
   def test_mutex_unlock_on_trap
     assert_in_out_err([], <<-INPUT, %w(locked unlocked false), [])
-      m = Mutex.new
+      m = Thread::Mutex.new
 
       trapped = false
       Signal.trap("INT") { |signo|
@@ -991,7 +1063,7 @@ q.pop
   def test_blocking_mutex_unlocked_on_fork
     bug8433 = '[ruby-core:55102] [Bug #8433]'
 
-    mutex = Mutex.new
+    mutex = Thread::Mutex.new
     flag = false
     mutex.lock
 
@@ -1050,7 +1122,8 @@ q.pop
   end
 
   def test_thread_name
-    t = Thread.start {}
+    t = Thread.start {sleep}
+    sleep 0.001 until t.stop?
     assert_nil t.name
     s = t.inspect
     t.name = 'foo'
@@ -1081,4 +1154,10 @@ q.pop
     t.kill
     t.join
   end
+
+  def test_thread_setname_in_initialize
+    bug12290 = '[ruby-core:74963] [Bug #12290]'
+    c = Class.new(Thread) {def initialize() self.name = "foo"; super; end}
+    assert_equal("foo", c.new {Thread.current.name}.value, bug12290)
+  end
 end
